@@ -31,7 +31,7 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 #include <sys/types.h>
 
 #include <sys/stat.h>
-#include <unistd.h>
+#include <stdlib.h>
 #include <syslog.h>
 
 #include "config_names.h"
@@ -363,18 +363,91 @@ connkill_cmd* read_connkill_cmd(app_private *ap)
 	return cmd;
 }
 
-void run_exec(connkill_cmd *cmd, char *ip, app_private *ap)
+static inline char *discard_const(char const* ip) { return (char*)ip;}
+
+int system_exec(connkill_cmd *cmd, char const *ip, app_private *ap)
 {
-	cmd->argv[cmd->ip_ndx] = ip;
+	assert(cmd);
+	assert(cmd->argv);
+	assert(ip);
+
+	cmd->argv[cmd->ip_ndx] = discard_const(ip);
+	unsigned needed = cmd->argc + 1;  // spaces + term. 0
+	for (unsigned i = 0; i < cmd->argc; ++i)
+		if (cmd->argv[i])
+			needed += strlen(cmd->argv[i]);
+
+	char command[needed];
+	char *p = &command[0];
+	for (unsigned i = 0; i < cmd->argc; ++i)
+		if (cmd->argv[i])
+			p += sprintf(p, "%s ", cmd->argv[i]);
+	*p = 0;
+
+#if defined TEST_MAIN
+	printf("running \"%s\"\n", command);
+#endif
+
+	int rtc = system(command);
+	if (rtc == -1)
+		report_error(ap, LOG_CRIT, "cannot run \"%s\": %s\n",
+			command, strerror(errno));
+
+	return rtc;
+}
+
+void run_exec(connkill_cmd *cmd, char const *ip, app_private *ap)
+{
+	assert(cmd);
+	assert(ip);
+
+	cmd->argv[cmd->ip_ndx] = discard_const(ip);
 	execv(cmd->argv[0], &cmd->argv[0]);
 	report_error(ap, LOG_CRIT, "cannot run %s: %s\n",
 		cmd->argv[0], strerror(errno));
+	_exit(1);
 }
 
+
 #if defined TEST_MAIN
+
+void display_result(connkill_cmd *cmd, int rtc)
+{
+	char const *result;
+	char buf[40];
+	if (rtc == -1)
+	{
+		result = "couldn't run";
+	}
+	else if (rtc)
+	{
+		if (WIFSIGNALED(rtc))
+		{
+			result = "was signaled";
+		}
+		else
+		{
+			int r = WEXITSTATUS(rtc);
+			if (r == 127)
+			{
+				result = "exited with status 127, no shell?";
+			}
+			else
+			{
+				sprintf(buf, "exited with status %d", r);
+				result = buf;
+			}
+		}
+	}
+	else
+		result = "exited successfully";
+
+	printf("%s %s\n", cmd->argv[0]? cmd->argv[0]: "NULL", result);
+}
+
 int main(int argc, char *argv[])
 {
-	int verbose = 0;
+	int verbose = 0, run4 = 0, run6 = 0;
 	app_private ap;
 	ap.mode = error_report_stderr;
 	ap.err_prefix = "TESTexec";
@@ -392,9 +465,11 @@ int main(int argc, char *argv[])
 		if (strcmp(a, "--help") == 0)
 		{
 			printf("%s: [-v] file...\n"
-			"Displays how file is parsed and which argument (ip_ndx) gets the IP\n"
+			"Displays how file is parsed and which argument (ip_ndx) gets the IP.\n"
 			"File(s) must be owned by the current user/group rather than root\n"
-			"Option -v also displays argument numbers; --version works\n",
+			"Option -v also displays argument numbers; --version works\n"
+			"option -r runs the command with argument 127.0.0.1,\n"
+			"option -R runs the command with argument ::1.\n",
 				argv[0]);
 			return 0;
 		}
@@ -404,19 +479,44 @@ int main(int argc, char *argv[])
 			verbose = 1;
 			continue;
 		}
-		
+
+		if (strcmp(a, "-r") == 0)
+		{
+			run4 = 1;
+			continue;
+		}
+
+		if (strcmp(a, "-R") == 0)
+		{
+			run6 = 1;
+			continue;
+		}
+
 		fname = a;
 		connkill_cmd *cmd = read_connkill_cmd(&ap);
 		puts(fname);
 		if (cmd)
 		{
-			printf("ip_ndx: %zu\nargc: %zu\n", cmd->ip_ndx, cmd->argc);
-			for (size_t j = 0; j <= cmd->argc; ++j)
+			printf("ip_ndx: %u\nargc: %u\n", cmd->ip_ndx, cmd->argc);
+			for (unsigned j = 0; j <= cmd->argc; ++j)
 			{
 				if (verbose)
-					printf("%zd: ", j);
+					printf("%u: ", j);
 				puts(cmd->argv[j]? cmd->argv[j]: "(NULL)");
 			}
+
+			if (run4)
+			{
+				int rtc = system_exec(cmd, "127.0.0.1", &ap);
+				display_result(cmd, rtc);
+			}
+
+			if (run6)
+			{
+				int rtc = system_exec(cmd, "::1", &ap);
+				display_result(cmd, rtc);
+			}
+
 			free(cmd);
 		}
 	}
