@@ -2,7 +2,7 @@
 * Ipq Berkeley db Daemon  BAN - ibd-ban
 * written by ale in milano on 13 sep 2008
 
-Copyright (C) 2008-2021 Alessandro Vesely
+Copyright (C) 2008-2023 Alessandro Vesely
 
 This file is part of Ipqbdb.
 
@@ -20,35 +20,16 @@ You should have received a copy of the GNU General Public License
 along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 
 */
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <math.h>
-#include <unistd.h>
-#include <limits.h>
-#include <signal.h>
 #include <syslog.h>
-
-#include <sys/types.h>
-
-// Berkeley DB v4.4
-#include <db.h>
-
 #include <popt.h>
 
 // format of data packet
 #include "config_names.h"
 #include "dbstruct.h"
 
-#include <time.h>
-
-// for inet_ntoa
-#define _GNU_SOURCE 1
-#include <sys/socket.h>
-#include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include "ipv4_util2.h"
+#include "ip_util.h"
 #include "initial_count.h"
 #include "percent_prob.h"
 #include "ext_exec.h"
@@ -58,20 +39,16 @@ static app_private ap;
 static char const err_prefix[] = "ibd-ban";
 #include "logline_written_block.h"
 
-#if !defined NDEBUG
-#define TRACE(x...) while (ap.mode == 0) { fprintf(stderr, ## x); break; }
-#else
-#define TRACE(x...)
-#endif
-
 #include "setsig_func.h"
+#include <assert.h>
 
 static int initial_count = IPQBDB_INITIAL_COUNT;
 static double initial_decay = IPQBDB_INITIAL_DECAY;
+static double index_factor = IPQBDB_INDEX_FACTOR;
 static char *db_block_name = IPQBDB_DATABASE_NAME;
 static char *db_white_name = IPQBDB_WHITE_DATABASE_NAME;
 static char *db_descr_name = IPQBDB_DESCR_DATABASE_NAME;
-static char *reason_string = "0";
+static char *reason_string = "n.a.";
 static char **ip_address;
 static int version_opt, help_opt, syslog_opt, cleanup_opt, verbose = -1;
 static int force_probability_opt, force_decay_opt, force_reason_opt;
@@ -80,7 +57,7 @@ static int exec_connkill_opt;
 static struct poptOption opttab[] =
 {
 	{"ip-addr", 'i', POPT_ARG_ARGV, &ip_address, 0,
-	"An address to ban, repeatable", "ip_addr"},
+	"Address or range to ban, repeatable option", "ip_addr"},
 	{"reason", 'r', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &reason_string, 0,
 	"The reason string (or its record number) for the block record", "int-or-string"},
 	{"force-reason", '\0', POPT_ARG_NONE, &force_reason_opt, 0,
@@ -93,6 +70,8 @@ static struct poptOption opttab[] =
 	"Set initial probability count", "integer"},
 	{"force-probability", '\0', POPT_ARG_NONE, &force_probability_opt, 0,
 	"Change the probability even if the current value is higher", NULL},
+	{"index-factor", 'f', POPT_ARG_DOUBLE|POPT_ARGFLAG_SHOW_DEFAULT, &index_factor, 0,
+	"Set range index factor", "float"},
 	{"log-category", '\0', POPT_ARG_NONE, &force_add_category_opt, 0,
 	"Enable logging the report category", NULL},
 	{"exec-connkill", 'e', POPT_ARG_NONE, &exec_connkill_opt, 0,
@@ -106,7 +85,7 @@ static struct poptOption opttab[] =
 	{"help", 'h', POPT_ARG_NONE, &help_opt, 0,
 	"This help.", NULL},
 	{"db-block", 'b', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &db_block_name, 0,
-	"The database of blocked IPv4 addresses", "filename"},
+	"The database of blocked addresses", "filename"},
 	{"db-white", 'w', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &db_white_name, 0,
 	"The whitelist database.", "filename"},
 	{"db-descr", 'd', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &db_descr_name, 0,
@@ -116,23 +95,22 @@ static struct poptOption opttab[] =
 	POPT_TABLEEND
 };
 
-static int get_ip_address(char *an_ip_address, uint32_t* ip_addr)
+typedef enum have_addr
 {
-	struct in_addr in;
-	int rtc = my_inet_pton(an_ip_address, &in);
-	if (rtc == 0 || in.s_addr == INADDR_ANY) // 0 is not valid
-	{
-		report_error(&ap, LOG_ERR, "invalid ip_address %s\n", an_ip_address);
-		return 1;
-	}
+	have_no_addr,
+	have_ipv4_addr,
+	have_ipv6_addr,
+	have_both_addr,
+	have_error_addr
+} have_addr;
 
-	memcpy(ip_addr, &in, sizeof *ip_addr);
-	return 0;
-}
-
-static int get_all_ip_address(uint32_t **ip_addr)
+static int get_all_ip_address(ip_range **ip_addr)
+/*
+* Read all addresses in a NULL-terminated array.  Return -1 in case of
+* error, NULL if no address is valid, or a have_addr flag.
+*/
 {
-	size_t count = 0;
+	unsigned count = 0;
 	char **a = ip_address;
 	if (a && *a)
 		while (*a++)
@@ -144,48 +122,145 @@ static int get_all_ip_address(uint32_t **ip_addr)
 		return 0;
 	}
 
-	uint32_t *addr = *ip_addr = calloc(count + 1, sizeof(uint32_t));
+	ip_range *addr = *ip_addr = calloc(count + 1, sizeof(ip_range));
 	if (addr == NULL)
 		return -1;
 
-	for (a = ip_address; *a; ++a, ++addr)
-		if (get_ip_address(*a, addr))
-			return -1;
+	addr[count].ip = -1; // sentinel
 
-	return 0;
+	int collect = 0, good = 0;
+	for (a = ip_address; *a; ++a, ++addr)
+	{
+		int err;
+		if ((err = parse_ip_address(*a, addr, NULL)) != 0)
+		{
+			report_error(&ap, LOG_ERR,
+				"invalid ip_address %s (%s)\n", *a,
+				parse_ip_invalid_what(err));
+			collect |= have_error_addr;
+		}
+		else if (addr->ip)
+		{
+			++good;
+			if (addr->ip == 4)
+			{
+				if (addr->args > 1)
+				{
+					--good;
+					report_error(&ap, LOG_ERR,
+						"invalid range %s (IPv4 range not supported)\n",
+						*a);
+					collect |= have_error_addr;
+					addr->ip = 0;
+				}
+				else
+					collect |= have_ipv4_addr;
+			}
+			else
+			{
+				assert(addr->ip == 6);
+				int invalid = 0;
+				if (addr->args > 1) // range
+				{
+					unsigned char ipv6[16];
+					int plen = addr->plen;
+					if (plen == 0)
+					{
+						memcpy(ipv6, addr->u2.ipv6, sizeof ipv6);
+						plen = range_ip(addr->u.ipv6, ipv6, 16);
+						if (memcmp(ipv6, addr->u2.ipv6, sizeof ipv6) != 0)
+							++invalid;
+						else
+							addr->plen = plen;
+					}
+
+					if (plen >= 128)
+						++invalid;
+
+					memcpy(ipv6, addr->u.ipv6, sizeof ipv6);
+					first_in_range(ipv6, plen);
+					if (memcmp(ipv6, addr->u.ipv6, sizeof ipv6) != 0)
+						++invalid;
+					if (invalid)
+					{
+						--good;
+						report_error(&ap, LOG_ERR,
+							"invalid range %s (not valid /%d)\n",
+							*a, plen);
+						collect |= have_error_addr;
+						addr->ip = 0;
+					}
+				}
+
+				if (!invalid)
+					collect |= have_ipv6_addr;
+			}
+		}
+	}
+
+	if (good == 0)
+	{
+		free(*ip_addr);
+		*ip_addr = NULL;
+	}
+	return collect;
 }
 
-static int display_check_whitelist(DB *db, uint32_t ip_addr, double *decay)
+static int display_check_whitelist(DB *db, ip_range *ip_addr, double *decay)
+/*
+* Return 0 if addr is to be blocked, 1 otherwise; < 0 on error.
+*
+* For IPv6 ranges, consider ip_addr to be whitelisted if it is fully
+* contained in the range; 
+*/
 {
-	int const rtc = check_whitelist(db, ip_addr, decay);
-	int rrc = rtc;
-	if (rrc == 1)
+	assert(db);
+	assert(ip_addr);
+	assert(decay);
+
+	unsigned char end_range[16];
+	int rtc = check_whitelist_e(db, base_ip_u(ip_addr), decay, end_range);
+	if (rtc < 0)
+		return rtc;
+
+	const int is_range = ip_addr->args > 1;
+	if (rtc == 1 && is_range) // found range
+	{
+		assert(ip_addr->ip == 6);
+		if (memcmp(end_range, ip_addr->u2.ipv6, sizeof end_range) < 0)
+			rtc = 0; // not fully contained
+	}
+
+	if (rtc == 0)
 	{
-		rrc = 0;
 		*decay = initial_decay;
 	}
 
 	if (verbose)
 	{
 		double const d = *decay;
-		struct in_addr in;
-		memcpy(&in, &ip_addr, sizeof in);
-		if (rtc == 0)
+		char addr[INET_RANGESTRLEN];
+
+		if (rtc == 1)
 		{
-			printf("%s is whitelisted", inet_ntoa(in));
+			printf("%s is whitelisted", is_range?
+				snprint_range(addr, sizeof addr, ip_addr):
+				my_inet_ntop(base_ip_u(ip_addr), addr));
 			if (d > 0)
 				printf(" with initial decay of %g seconds\n", d);
 			else
 				fputs(" and won't be blocked\n", stdout);
 		}
-		else if (rtc == 1)
+		else if (rtc == 0)
 		{
 			printf("%s is not whitelisted, initial decay of %g seconds\n",
-				inet_ntoa(in), d);
+				is_range?
+				snprint_range(addr, sizeof addr, ip_addr):
+				my_inet_ntop(base_ip_u(ip_addr), addr), d);
 		}
 	}
 
-	return rrc;
+	return rtc;
 }
 
 static const char *month_abbr(unsigned ndx)
@@ -227,25 +302,46 @@ static void tab_printtime(char const *hdr, time_t ref, time_t now)
 	printf("\t%-15s %18s %29s\n", hdr, buf2, buf);
 }
 
-static void display_written_block(uint32_t ip_addr, int reason_id,
-	int write_block_force,
+#define INET6_RANGESTRLEN (INET6_ADDRSTRLEN + 5)
+static const char *display_range(ip_u *ip_addr, int plen, char addr[INET6_RANGESTRLEN])
+{
+	if (ip_addr->ip == 6 && plen > 0 && plen < 128)
+	{
+		unsigned char rng[16];
+		memcpy(rng, ip_addr->u.ipv6, sizeof rng);
+		first_in_range(rng, plen);
+		sprintf(addr, "%s/%d", inet_ntop(AF_INET6, rng, addr, INET6_RANGESTRLEN), plen);
+		return addr;
+	}
+	else
+		return my_inet_ntop(ip_addr, addr);
+}
+
+static void display_written_block(ip_u *ip_addr, void *key_data,
+	int reason_id, int write_block_force,
 	ip_data_t *ip_data, ip_data_t *old_data, DB *db_d)
 {
 	// if syslog_opt, log a line anyway (unlikely given -v).
 	if (ap.mode)
-		logline_written_block(ip_addr, reason_id,
+		logline_written_block(ip_addr, key_data, reason_id,
 			write_block_force, ip_data, old_data, db_d);
 
-	struct in_addr in;
+	char addr[INET6_RANGESTRLEN];
+	display_range(ip_addr, old_data? old_data->plen: ip_data->plen, addr);
 	char descr[19 /* IPQBDB_DESCR_MAX_LENGTH */];
-	memcpy(&in, &ip_addr, sizeof in);
 	time_t now = ip_data->last_update;
 
 	get_descr(db_d, ip_data->reason_id,
 		get_descr_quoted | get_descr_add_id, descr, sizeof descr);
 	if (old_data)
 	{
-		printf("changed probability record for %s", inet_ntoa(in));
+		printf("changed probability record for %s", addr);
+		if (ip_addr->ip == 6 &&
+			memcmp(ip_addr->u.ipv6, key_data, sizeof ip_addr->u.ipv6))
+		{
+			display_range(ip_addr, ip_data->plen, addr);
+			printf("->%s", addr);
+		}
 		if (ip_data->reason_id == old_data->reason_id)
 			printf(", %s", descr);
 		fputc('\n', stdout);
@@ -271,7 +367,7 @@ static void display_written_block(uint32_t ip_addr, int reason_id,
 	else
 	{
 		printf("inserted probability record for %s (reason %d)\n",
-			inet_ntoa(in), ip_data->reason_id);
+			addr, ip_data->reason_id);
 		printf("\tdecay: %g, probability: %d=%.2f%%\n",
 			ip_data->decay,
 			ip_data->probability, percent_prob(ip_data->probability));
@@ -279,14 +375,18 @@ static void display_written_block(uint32_t ip_addr, int reason_id,
 }
 
 static int
-write_block_display(DB *db, uint32_t ip_addr, double decay, int reason_id, DB *db_d)
+write_block_display(DB *db, ip_range *ip_addr, double decay, int reason_id, DB *db_d)
 /*
 * Do initial_count and force options, then write and possibly display block.
 * Return -1 on error, 0 if all ok, 1 if all ok and probability reached 100%.
 */
 {
-	int probability;
+	assert(db);
+	assert(ip_addr);
+	assert(ip_addr->ip == 4 || ip_addr->ip == 6);
+	assert(ip_addr->args == 1 || ip_addr->ip == 6); // range only IPv6
 
+	int probability;
 	if (initial_count == 999)
 		probability = 0;
 	else
@@ -306,29 +406,44 @@ write_block_display(DB *db, uint32_t ip_addr, double decay, int reason_id, DB *d
 	if (force_reason_opt) force |= write_block_force_reason;
 	if (force_add_category_opt) force |= write_block_force_add_category;
 
-	if (write_block(db, ip_addr, decay, reason_id, &probability, force,
-		verbose? display_written_block:
-		ap.mode? logline_written_block: NULL, db_d))
-			return -1;
+	if (write_block(db, base_ip_u(ip_addr), decay, reason_id,
+			&probability, force, verbose? display_written_block:
+			ap.mode? logline_written_block: NULL, db_d))
+				return -1;
 
 	return probability >= RAND_MAX;
 }
 
-static void run_exec_ip(connkill_cmd *cmd, uint32_t ip_addr)
+static int system_exec_ip(connkill_cmd *cmd, ip_u *ip_addr)
 {
-	char *my_ip_address;
-	struct in_addr in;
-	memcpy(&in, &ip_addr, sizeof in);
+	assert(cmd);
+	assert(ip_addr);
+	assert(ip_addr->ip == 4 || ip_addr->ip == 6);
 
-	my_ip_address = strdup(inet_ntoa(in));
-	run_exec(cmd, my_ip_address, &ap);
+	char addr[INET6_ADDRSTRLEN];
+	char const *a = my_inet_ntop(ip_addr, addr);
+
+	return system_exec(cmd, a, &ap);
+}
+
+static void run_exec_ip(connkill_cmd *cmd, ip_u *ip_addr)
+{
+	assert(cmd);
+	assert(ip_addr);
+	assert(ip_addr->ip == 4 || ip_addr->ip == 6);
+
+	char addr[INET6_ADDRSTRLEN];
+	char const *a = my_inet_ntop(ip_addr, addr);
+ 
+	run_exec(cmd, a, &ap);
 }
 
 int main(int argc, char const *argv[])
 {
 	static const char optaliases[] = IPQBDB_OPTION_FILE;
 	int rtc = 0, errs = 0;
-	uint32_t *ip_addr = NULL;
+	ip_range *ip_addr = NULL;
+	int collect_have_addr = 0;
 
 	poptContext opt = poptGetContext(err_prefix, argc, argv, opttab, 0);
 
@@ -385,7 +500,8 @@ int main(int argc, char const *argv[])
 			errs = 1;
 		}
 
-		if (get_all_ip_address(&ip_addr))
+		collect_have_addr = get_all_ip_address(&ip_addr);
+		if (collect_have_addr < 0 || ip_addr == NULL)
 			errs = 3;
 
 		if (initial_decay < 0.0)
@@ -394,6 +510,9 @@ int main(int argc, char const *argv[])
 				initial_decay);
 			errs = 3;
 		}
+
+		if (index_factor != IPQBDB_INDEX_FACTOR)
+			set_ipqbdb_index_factor(index_factor);
 	}
 
 	if (errs == 1 && !syslog_opt)
@@ -405,9 +524,9 @@ int main(int argc, char const *argv[])
 		rtc = 1;
 	else
 	{
-		char *block_fname = NULL;
-		char *descr_fname = NULL;
-		char *white_fname = NULL;
+		switchable_fname *block_fname = NULL;
+		switchable_fname *descr_fname = NULL;
+		switchable_fname *white_fname = NULL;
 
 		int reason_id = -1;
 		char *t = NULL;
@@ -445,18 +564,26 @@ int main(int argc, char const *argv[])
 		{
 			DB_ENV *db_env = NULL;
 			DB *db_block = NULL, *db_white = NULL, *db_de1 = NULL, *db_de2 = NULL;
-			uint32_t exec_connkill = 0;
+			DB *db6_block = NULL, *db6_white = NULL;
+			DBC *db6c_white = NULL;
+			int exec_connkill_cnt = 0;
 
 			setsigs();
 
-			rtc = open_database(block_fname? block_fname: descr_fname,
-					&ap, &db_env, block_fname? &db_block: NULL);
+			rtc = open_database(block_fname, &ap, &db_env,
+					collect_have_addr & have_ipv4_addr? &db_block: NULL,
+					collect_have_addr & have_ipv6_addr? &db6_block: NULL);
 
 			if (rtc == 0 && white_fname && caught_signal == 0)
-				rtc = open_database(white_fname, &ap, &db_env, &db_white);
+				rtc = open_database(white_fname, &ap, &db_env,
+					collect_have_addr & have_ipv4_addr? &db_white: NULL,
+					collect_have_addr & have_ipv6_addr? &db6_white: NULL);
+
+			if (rtc == 0 && db6_white && caught_signal == 0)
+				rtc = db6_white->cursor(db6_white, NULL, &db6c_white, 0);
 
 			if (rtc == 0 && descr_fname && caught_signal == 0)
-				rtc = open_descrdb(descr_fname, db_env, &db_de1, &db_de2);
+				rtc = open_descrdb(descr_fname->fname, db_env, &db_de1, &db_de2);
 
 			if (rtc == 0 && reason_id == -1 && caught_signal == 0)
 				rtc = get_reason_id(&reason_id, reason_string,
@@ -465,19 +592,40 @@ int main(int argc, char const *argv[])
 			if (rtc == 0 && ip_addr && caught_signal == 0)
 			{
 				double decay;
-				uint32_t *next_ip = ip_addr;
-				while (rtc == 0 && *next_ip && caught_signal == 0)
+				ip_range *next_ip = ip_addr;
+				while (rtc == 0 && next_ip->ip >= 0 && caught_signal == 0)
 				{
-					rtc = display_check_whitelist(db_white, *next_ip, &decay);
-
-					if (rtc == 0 && caught_signal == 0 && decay > 0.0)
+					int ip = next_ip->ip;
+					if (ip > 0)
 					{
-						int a_rc = write_block_display(
-							db_block, *next_ip, decay, reason_id, db_de1);
-						if (a_rc > 0)
-							exec_connkill = *next_ip;
-						else if (a_rc != 0)
-							rtc = -1; // bail out on error
+						assert(ip == 4 || ip == 6);
+						void *db_or_c;
+						if (ip == 4)
+							db_or_c = db_white;
+						else
+							db_or_c = db6c_white;
+						rtc = display_check_whitelist(db_or_c, next_ip, &decay);
+						if (rtc < 0)
+						{
+							DB *db_err = ip == 4? db_white: db6_white;
+							db_err->errx(db_err, "cannot check whitelist");
+						}
+
+						if (rtc == 0 && caught_signal == 0 && decay > 0.0)
+						{
+							int a_rc = write_block_display(
+								ip == 4? db_block: db6_block,
+								next_ip, decay, reason_id, db_de1);
+							if (a_rc > 0)
+							{
+								exec_connkill_cnt += 1;
+								next_ip->kill = 1;
+							}
+							else if (a_rc != 0)
+								rtc = -1; // bail out on error
+						}
+						else if (rtc == 1) // whitelisted
+							rtc = 0;
 					}
 					++next_ip;
 				}
@@ -490,20 +638,66 @@ int main(int argc, char const *argv[])
 			close_db(db_de1);
 			close_db(db_white);
 			close_db(db_block);
+			if (db6c_white)
+				db6c_white->close(db6c_white);
+			close_db(db6_white);
+			close_db(db6_block);
 
 			close_dbenv(db_env, cleanup_opt);
 
-			if (exec_connkill && exec_connkill_opt)
+			if (exec_connkill_cnt > 0 && exec_connkill_opt && caught_signal == 0)
 			{
-				connkill_cmd *cmd = read_connkill_cmd(&ap);
+				connkill_cmd *const cmd = read_connkill_cmd(&ap);
+				ip_range *next_ip = ip_addr;
 				if (cmd)
-					run_exec_ip(cmd, exec_connkill);
+				{
+					while (next_ip->ip >= 0 && caught_signal == 0)
+					{
+						int ip = next_ip->ip;
+						if (ip > 0 && next_ip->kill)
+						{
+							if (exec_connkill_cnt == 1)
+								run_exec_ip(cmd, base_ip_u(next_ip)); // won't return
 
-				rtc = 2; // either exec or cmd failed
+							int rtc_kill = system_exec_ip(cmd, base_ip_u(next_ip));
+							if (rtc_kill == -1)
+							{
+								rtc = 2;
+								break;
+							}
+
+							if (rtc_kill)
+							{
+								if (WIFSIGNALED(rtc_kill))
+								{
+									rtc = 2;
+									caught_signal = 1;
+								}
+								else
+								{
+									int r = WEXITSTATUS(rtc_kill);
+									if (r == 127)
+									{
+										rtc = 2;
+										break;
+									}
+
+									if (r > rtc)
+										rtc = r;
+								}
+							}
+						}
+						++next_ip;
+					}
+				}
 			}
 		}
+		free(block_fname);
+		free(descr_fname);
+		free(white_fname);
 	}
 
+	free(ip_addr);
 	return rtc;
 }
 
