@@ -805,7 +805,7 @@ do_del(DB *db, DB*db_d, ip_range* ip, time_t *secs, int *prob, histog *his)
 
 			if (p == NULL)
 			{
-				rtc = DB_NOTFOUND;
+				//rtc = DB_NOTFOUND;
 				break;
 			}
 
@@ -1440,7 +1440,7 @@ int main(int argc, char const *argv[])
 								" (%jd for IPv6)", (intmax_t)secs_6);
 					}
 					else
-						range[0] = 0;
+						secs_six[0] = 0;
 
 					printf("%s%s%s records in IP range %s\n"
 						"\t%s%smin age %jd secs%s, max age %jd secs,"
