@@ -1,7 +1,7 @@
 /*
 * initial_count.h
 
-Copyright (C) 2011 Alessandro Vesely
+Copyright (C) 2011-2019 Alessandro Vesely
 
 This file is part of Ipqbdb.
 
@@ -25,6 +25,15 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 #include <limits.h> // CHAR_BIT
 #include <stdlib.h> // RAND_MAX
 
+//#if defined HAVE_CONFIG
+#include <config.h>
+//#endif
+
+
+#if ! defined IPQBDB_INITIAL_COUNT
+#define IPQBDB_INITIAL_COUNT 5
+#endif
+
 static inline int count2prob(unsigned initial_count)
 // return a non-negative integer, 0 if count is too high
 // the returned value should be always increased by one anyway, for rounding
